@@ -2,11 +2,13 @@ import asyncio
 import contextlib
 import logging
 import re
+import time
+from typing import Dict, Optional, Tuple
 
 import httpx
 import jwt
 
-from ._auth_provider import get_auth_provider
+from ._auth_provider import cleanup_shared_keys, get_auth_provider
 from ._blob_client import BlobClient
 from ._decorators import (
     raise_for_status,
@@ -24,16 +26,21 @@ DEFAULT_TIMEOUT = httpx.Timeout(30.0)
 class SumoClient:
     """Authenticate and perform requests to the Sumo API."""
 
+    _client: httpx.Client
+    _async_client: httpx.AsyncClient
+
     def __init__(
         self,
         env: str,
-        token: str = None,
-        interactive: bool = False,
+        token: Optional[str] = None,
+        interactive: bool = True,
         devicecode: bool = False,
         verbosity: str = "CRITICAL",
         retry_strategy=RetryStrategy(),
         timeout=DEFAULT_TIMEOUT,
         case_uuid=None,
+        http_client=None,
+        async_http_client=None,
     ):
         """Initialize a new Sumo object
 
@@ -54,8 +61,19 @@ class SumoClient:
         self._verbosity = verbosity
 
         self._retry_strategy = retry_strategy
-        self._client = httpx.Client()
-        self._async_client = httpx.AsyncClient()
+        if http_client is None:
+            self._client = httpx.Client()
+            self._borrowed_client = False
+        else:
+            self._client = http_client
+            self._borrowed_client = True
+
+        if async_http_client is None:
+            self._async_client = httpx.AsyncClient()
+            self._borrowed_async_client = False
+        else:
+            self._async_client = async_http_client
+            self._borrowed_async_client = True
 
         self._timeout = timeout
 
@@ -81,6 +99,9 @@ class SumoClient:
                 refresh_token = token
                 pass
             pass
+
+        cleanup_shared_keys()
+
         self.auth = get_auth_provider(
             client_id=APP_REGISTRATION[env]["CLIENT_ID"],
             authority=f"{AUTHORITY_HOST_URI}/{TENANT_ID}",
@@ -92,35 +113,35 @@ class SumoClient:
             case_uuid=case_uuid,
         )
 
-        if env == "localhost":
+        if env == "prod":
+            self.base_url = "https://api.sumo.equinor.com/api/v1"
+        elif env == "localhost":
             self.base_url = "http://localhost:8084/api/v1"
         else:
             self.base_url = f"https://main-sumo-{env}.radix.equinor.com/api/v1"
-            pass
         return
 
     def __enter__(self):
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
-        self._client.close()
-        self._client = None
+    def __exit__(self, *_):
+        if not self._borrowed_client:
+            self._client.close()
         return False
 
     async def __aenter__(self):
         return self
 
-    async def __aexit__(self, exc_type, exc_value, traceback):
-        await self._async_client.aclose()
-        self._async_client = None
+    async def __aexit__(self, *_):
+        if not self._borrowed_async_client:
+            await self._async_client.aclose()
         return False
 
     def __del__(self):
-        if self._client is not None:
+        if self._client is not None and not self._borrowed_client:
             self._client.close()
-            self._client = None
             pass
-        if self._async_client is not None:
+        if self._async_client is not None and not self._borrowed_async_client:
 
             async def closeit(client):
                 await client.aclose()
@@ -131,7 +152,6 @@ class SumoClient:
                 loop.create_task(closeit(self._async_client))
             except RuntimeError:
                 pass
-            self._async_client = None
             pass
 
     def authenticate(self):
@@ -167,7 +187,7 @@ class SumoClient:
         )
 
     @raise_for_status
-    def get(self, path: str, params: dict = None) -> dict:
+    def get(self, path: str, params: Optional[Dict] = None) -> httpx.Response:
         """Performs a GET-request to the Sumo API.
 
         Args:
@@ -208,6 +228,11 @@ class SumoClient:
                 path,
             )
             is not None
+            or re.match(
+                r"^/tasks\('[0-9a-fA-F-]{8}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{12}'\)/result$",  # noqa: E501
+                path,
+            )
+            is not None
         ):
             follow_redirects = True
 
@@ -228,9 +253,9 @@ class SumoClient:
     def post(
         self,
         path: str,
-        blob: bytes = None,
-        json: dict = None,
-        params: dict = None,
+        blob: Optional[bytes] = None,
+        json: Optional[dict] = None,
+        params: Optional[dict] = None,
     ) -> httpx.Response:
         """Performs a POST-request to the Sumo API.
 
@@ -301,7 +326,10 @@ class SumoClient:
 
     @raise_for_status
     def put(
-        self, path: str, blob: bytes = None, json: dict = None
+        self,
+        path: str,
+        blob: Optional[bytes] = None,
+        json: Optional[dict] = None,
     ) -> httpx.Response:
         """Performs a PUT-request to the Sumo API.
 
@@ -346,7 +374,9 @@ class SumoClient:
         return retryer(_put)
 
     @raise_for_status
-    def delete(self, path: str, params: dict = None) -> dict:
+    def delete(
+        self, path: str, params: Optional[dict] = None
+    ) -> httpx.Response:
         """Performs a DELETE-request to the Sumo API.
 
         Args:
@@ -354,7 +384,7 @@ class SumoClient:
             params: query parameters, as dictionary
 
         Returns:
-            Sumo JSON resposne as a dictionary
+            Sumo JSON response as a dictionary
 
         Examples:
             Deleting object::
@@ -383,6 +413,45 @@ class SumoClient:
 
         return retryer(_delete)
 
+    def _get_retry_details(self, response_in) -> Tuple[str, int]:
+        assert response_in.status_code == 202, (
+            "Incorrect status code; expcted 202"
+        )
+        headers = response_in.headers
+        location: str = headers.get("location")
+        assert location is not None, "Missing header: Location"
+        assert location.startswith(self.base_url)
+        retry_after = headers.get("retry-after")
+        assert retry_after is not None, "Missing header: Retry-After"
+        location = location[len(self.base_url) :]
+        retry_after = int(retry_after)
+        return location, retry_after
+
+    def poll(
+        self, response_in: httpx.Response, timeout=None
+    ) -> httpx.Response:
+        """Poll a specific endpoint until a result is obtained.
+
+        Args:
+            response_in: httpx.Response from a previous request, with 'location' and 'retry-after' headers.
+
+        Returns:
+            A new httpx.response object.
+        """
+        location, retry_after = self._get_retry_details(response_in)
+        expiry = time.time() + timeout if timeout is not None else None
+        while True:
+            time.sleep(retry_after)
+            response = self.get(location)
+            if response.status_code != 202:
+                return response
+            if expiry is not None and time.time() > expiry:
+                raise httpx.TimeoutException(
+                    "No response within specified timeout."
+                )
+            location, retry_after = self._get_retry_details(response)
+            pass
+
     def getLogger(self, name):
         """Gets a logger object that sends log objects into the message_log
         index for the Sumo instance.
@@ -424,16 +493,21 @@ class SumoClient:
     def client_for_case(self, case_uuid):
         """Instantiate and return new SumoClient for accessing the
         case identified by *case_uuid*."""
-        return SumoClient(
-            env=self.env,
-            verbosity=self._verbosity,
-            retry_strategy=self._retry_strategy,
-            timeout=self._timeout,
-            case_uuid=case_uuid,
-        )
+        if self.auth.has_case_token(case_uuid):
+            return SumoClient(
+                env=self.env,
+                verbosity=self._verbosity,
+                retry_strategy=self._retry_strategy,
+                timeout=self._timeout,
+                case_uuid=case_uuid,
+            )
+        else:
+            return self
 
     @raise_for_status_async
-    async def get_async(self, path: str, params: dict = None):
+    async def get_async(
+        self, path: str, params: Optional[dict] = None
+    ) -> httpx.Response:
         """Performs an async GET-request to the Sumo API.
 
         Args:
@@ -474,6 +548,11 @@ class SumoClient:
                 path,
             )
             is not None
+            or re.match(
+                r"^/tasks\('[0-9a-fA-F-]{8}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{4}-[0-9a-fA-F-]{12}'\)/result$",  # noqa: E501
+                path,
+            )
+            is not None
         ):
             follow_redirects = True
 
@@ -494,9 +573,9 @@ class SumoClient:
     async def post_async(
         self,
         path: str,
-        blob: bytes = None,
-        json: dict = None,
-        params: dict = None,
+        blob: Optional[bytes] = None,
+        json: Optional[dict] = None,
+        params: Optional[dict] = None,
     ) -> httpx.Response:
         """Performs an async POST-request to the Sumo API.
 
@@ -568,7 +647,10 @@ class SumoClient:
 
     @raise_for_status_async
     async def put_async(
-        self, path: str, blob: bytes = None, json: dict = None
+        self,
+        path: str,
+        blob: Optional[bytes] = None,
+        json: Optional[dict] = None,
     ) -> httpx.Response:
         """Performs an async PUT-request to the Sumo API.
 
@@ -613,7 +695,9 @@ class SumoClient:
         return await retryer(_put)
 
     @raise_for_status_async
-    async def delete_async(self, path: str, params: dict = None) -> dict:
+    async def delete_async(
+        self, path: str, params: Optional[dict] = None
+    ) -> httpx.Response:
         """Performs an async DELETE-request to the Sumo API.
 
         Args:
@@ -621,7 +705,7 @@ class SumoClient:
             params: query parameters, as dictionary
 
         Returns:
-            Sumo JSON resposne as a dictionary
+            Sumo JSON response as a dictionary
 
         Examples:
             Deleting object::
@@ -649,3 +733,28 @@ class SumoClient:
         retryer = self._retry_strategy.make_retryer_async()
 
         return await retryer(_delete)
+
+    async def poll_async(
+        self, response_in: httpx.Response, timeout=None
+    ) -> httpx.Response:
+        """Poll a specific endpoint until a result is obtained.
+
+        Args:
+            response_in: httpx.Response from a previous request, with 'location' and 'retry-after' headers.
+
+        Returns:
+            A new httpx.response object.
+        """
+        location, retry_after = self._get_retry_details(response_in)
+        expiry = time.time() + timeout if timeout is not None else None
+        while True:
+            await asyncio.sleep(retry_after)
+            response = await self.get_async(location)
+            if response.status_code != 202:
+                return response
+            if expiry is not None and time.time() > expiry:
+                raise httpx.TimeoutException(
+                    "No response within specified timeout."
+                )
+            location, retry_after = self._get_retry_details(response)
+            pass
